# Copyright (c) 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for the failover segment api."""

import ddt
import mock
from oslo_serialization import jsonutils
from six.moves import http_client as http
from webob import exc

from masakari.api.openstack.ha import segments
from masakari import exception
from masakari.ha import api as ha_api
from masakari.objects import base as obj_base
from masakari.objects import segment as segment_obj
from masakari import test
from masakari.tests.unit.api.openstack import fakes
from masakari.tests import uuidsentinel


def _make_segment_obj(segment_dict):
    return segment_obj.FailoverSegment(**segment_dict)


def _make_segments_list(segments_list):
    return segment_obj.FailoverSegment(objects=[
        _make_segment_obj(a) for a in segments_list])

FAILOVER_SEGMENT_LIST = [
    {"name": "segment1", "id": "1", "service_type": "COMPUTE",
     "recovery_method": "auto", "uuid": uuidsentinel.fake_segment,
     "description": "failover_segment for compute"},

    {"name": "segment2", "id": "2", "service_type": "CINDER",
     "recovery_method": "reserved_host", "uuid": uuidsentinel.fake_segment2,
     "description": "failover_segment for cinder"}
]

FAILOVER_SEGMENT_LIST = _make_segments_list(FAILOVER_SEGMENT_LIST)

FAILOVER_SEGMENT = {"name": "segment1", "id": "1",
                    "service_type": "COMPUTE", "recovery_method": "auto",
                    "uuid": uuidsentinel.fake_segment,
                    "description": "failover_segment for compute"}

FAILOVER_SEGMENT = _make_segment_obj(FAILOVER_SEGMENT)


@ddt.ddt
class FailoverSegmentTestCase(test.TestCase):
    """Test Case for failover segment api."""

    bad_request = exception.ValidationError

    def setUp(self):
        super(FailoverSegmentTestCase, self).setUp()
        self.controller = segments.SegmentsController()
        self.req = fakes.HTTPRequest.blank('/v1/segments',
                                           use_admin_context=True)
        self.context = self.req.environ['masakari.context']

    @property
    def app(self):
        return fakes.wsgi_app_v1(init_only='segments')

    def _assert_segment_data(self, expected, actual):
        self.assertTrue(obj_base.obj_equal_prims(expected, actual),
                        "The failover segment objects were not equal")

    @mock.patch.object(ha_api.FailoverSegmentAPI, 'get_all')
    def test_index(self, mock_get_all):

        mock_get_all.return_value = FAILOVER_SEGMENT_LIST

        result = self.controller.index(self.req)
        result = result['segments']
        self._assert_segment_data(FAILOVER_SEGMENT_LIST,
                                  _make_segments_list(result))

    @mock.patch.object(ha_api.FailoverSegmentAPI, 'get_all')
    def test_index_marker_not_found(self, mock_get_all):
        fake_request = fakes.HTTPRequest.blank('/v1/segments?marker=12345',
                                               use_admin_context=True)
        mock_get_all.side_effect = exception.MarkerNotFound(marker="12345")
        self.assertRaises(exc.HTTPBadRequest, self.controller.index,
                          fake_request)

    @ddt.data(
        # limit negative
        'limit=-1',

        # invalid sort key
        'sort_key=abcd',

        # invalid sort dir
        'sort_dir=abcd')
    def test_index_invalid(self, param):
        req = fakes.HTTPRequest.blank("/v1/segments?%s" % param,
                                      use_admin_context=True)

        self.assertRaises(exc.HTTPBadRequest, self.controller.index, req)

    @mock.patch.object(ha_api.FailoverSegmentAPI, 'create_segment')
    def test_create(self, mock_create):
        body = {
            "segment": {
                "name": "segment1",
                "service_type": "COMPUTE",
                "recovery_method": "auto",
                "description": "failover_segment for compute"
            }
        }
        mock_create.return_value = FAILOVER_SEGMENT
        result = self.controller.create(self.req, body=body)
        result = result['segment']
        self._assert_segment_data(FAILOVER_SEGMENT, _make_segment_obj(result))

    @mock.patch.object(ha_api.FailoverSegmentAPI, 'create_segment')
    def test_create_with_duplicate_segment_name(self, mock_create):
        body = {
            "segment": {
                "name": "segment1",
                "service_type": "COMPUTE",
                "recovery_method": "auto",
                "description": "failover_segment for compute"
            }
        }
        mock_create.side_effect = (exception.
                                   FailoverSegmentExists(name='segment1'))
        self.assertRaises(exc.HTTPConflict, self.controller.create,
                          self.req, body=body)

    @mock.patch('masakari.rpc.get_client')
    @mock.patch.object(ha_api.FailoverSegmentAPI, 'create_segment')
    def test_create_success_with_201_response_code(
        self, mock_client, mock_create):
        body = {
            "segment": {
                "name": "segment1",
                "service_type": "COMPUTE",
                "recovery_method": "auto",
                "description": "failover_segment for compute"
            }
        }
        fake_req = self.req
        fake_req.headers['Content-Type'] = 'application/json'
        fake_req.method = 'POST'
        fake_req.body = jsonutils.dump_as_bytes(body)
        resp = fake_req.get_response(self.app)
        self.assertEqual(http.CREATED, resp.status_code)

    @ddt.data(
        # no segment
        {"body": {
            "name": "segment1",
            "service_type": "COMPUTE",
            "recovery_method": "auto",
            "description": "failover_segment for compute"}},

        # no name
        {"body": {
            "segment": {
                "service_type": "COMPUTE",
                "recovery_method": "auto",
                "description": "failover_segment for compute"}}},

        # name with leading trailing spaces
        {"body": {
            "segment": {
                "name": "    segment1    ",
                "service_type": "COMPUTE",
                "recovery_method": "auto",
                "description": "failover_segment for compute"}}},

        # null name
        {"body": {
            "segment": {
                "name": "",
                "service_type": "COMPUTE",
                "recovery_method": "auto",
                "description": "failover_segment for compute"}}},

        # name too long
        {"body": {
            "segment": {
                "name": "segment1" * 255,
                "service_type": "COMPUTE",
                "recovery_method": "auto",
                "description": "failover_segment for compute"}}},

        # extra invalid args
        {"body": {
            "segment": {
                "name": "segment1" * 255,
                "service_type": "COMPUTE",
                "recovery_method": "auto",
                "description": "failover_segment for compute",
                "foo": "fake_foo"}}}
    )
    @ddt.unpack
    def test_create_failure(self, body):
        self.assertRaises(self.bad_request, self.controller.create,
                          self.req, body=body)

    @mock.patch.object(ha_api.FailoverSegmentAPI, 'get_segment')
    def test_show(self, mock_get_segment):

        mock_get_segment.return_value = FAILOVER_SEGMENT

        result = self.controller.show(self.req, uuidsentinel.fake_segment)
        result = result['segment']
        self._assert_segment_data(FAILOVER_SEGMENT, _make_segment_obj(result))

    @mock.patch.object(ha_api.FailoverSegmentAPI, 'get_segment')
    def test_show_with_non_existing_id(self, mock_get_segment):

        mock_get_segment.side_effect = exception.FailoverSegmentNotFound(
            id="2")
        self.assertRaises(exc.HTTPNotFound,
                          self.controller.show, self.req, "2")

    @ddt.data(
        {"body": {"segment": {"name": "segment1", "service_type": "COMPUTE",
                              "recovery_method": "auto"}}},

        # with name only
        {"body": {"segment": {"name": "segment1"}}}

    )
    @ddt.unpack
    @mock.patch.object(ha_api.FailoverSegmentAPI, 'update_segment')
    def test_update(self, mock_update_segment, body):
        mock_update_segment.return_value = FAILOVER_SEGMENT

        result = self.controller.update(self.req, uuidsentinel.fake_segment,
                                        body=body)

        result = result['segment']
        self._assert_segment_data(FAILOVER_SEGMENT, _make_segment_obj(result))

    @ddt.data(
        # no updates
        {"test_data": {"segment": {}}},

        # no update key
        {"test_data": {"asdf": {}}},

        # wrong updates
        {"test_data": {"segment": {"name": "disable", "foo": "bar"}}},

        # null name
        {"test_data": {"segment": {"name": ""}}},

        # name too long
        {"test_data": {"segment": {"name": "x" * 256}}}
    )
    @ddt.unpack
    def test_update_failure(self, test_data):
        self.assertRaises(self.bad_request, self.controller.update,
                          self.req, uuidsentinel.fake_segment, body=test_data)

    @mock.patch.object(ha_api.FailoverSegmentAPI, 'update_segment')
    def test_update_with_non_exising_segment(self, mock_update_segment):

        test_data = {"segment": {"name": "segment11"}}
        mock_update_segment.side_effect = exception.FailoverSegmentNotFound(
            id="2")
        self.assertRaises(exc.HTTPNotFound, self.controller.update,
                self.req, "2", body=test_data)

    @mock.patch.object(ha_api.FailoverSegmentAPI, 'update_segment')
    def test_update_with_duplicated_name(self, mock_update_segment):
        test_data = {"segment": {"name": "segment1"}}
        mock_update_segment.side_effect = exception.FailoverSegmentExists(
            name="segment1")
        self.assertRaises(exc.HTTPConflict, self.controller.update,
                self.req, uuidsentinel.fake_segment, body=test_data)

    @mock.patch.object(ha_api.FailoverSegmentAPI, 'delete_segment')
    def test_delete_segment(self, mock_delete):

        self.controller.delete(self.req, uuidsentinel.fake_segment)
        self.assertTrue(mock_delete.called)

    @mock.patch.object(ha_api.FailoverSegmentAPI, 'delete_segment')
    def test_delete_segment_not_found(self, mock_delete):

        mock_delete.side_effect = exception.FailoverSegmentNotFound(
            id=uuidsentinel.fake_segment)
        self.assertRaises(exc.HTTPNotFound, self.controller.delete,
                self.req, uuidsentinel.fake_segment)

    @mock.patch('masakari.rpc.get_client')
    @mock.patch.object(ha_api.FailoverSegmentAPI, 'delete_segment')
    def test_delete_segment_with_204_status(self, mock_client, mock_delete):
        url = '/v1/segments/%s' % uuidsentinel.fake_segment
        fake_req = fakes.HTTPRequest.blank(url, use_admin_context=True)
        fake_req.headers['Content-Type'] = 'application/json'
        fake_req.method = 'DELETE'
        resp = fake_req.get_response(self.app)
        self.assertEqual(http.NO_CONTENT, resp.status_code)


class FailoverSegmentTestCasePolicyNotAuthorized(test.NoDBTestCase):
    """Test Case for failover segment non admin."""

    def setUp(self):
        super(FailoverSegmentTestCasePolicyNotAuthorized, self).setUp()
        self.controller = segments.SegmentsController()
        self.req = fakes.HTTPRequest.blank('/v1/segments')
        self.context = self.req.environ['masakari.context']

    def _check_rule(self, exc, rule_name):
        self.assertEqual(
            "Policy doesn't allow %s to be performed." % rule_name,
            exc.format_message())

    def test_index_no_admin(self):
        rule_name = "os_masakari_api:segments:index"
        self.policy.set_rules({rule_name: "project:non_fake"})
        exc = self.assertRaises(exception.PolicyNotAuthorized,
                                self.controller.index,
                                self.req)
        self._check_rule(exc, rule_name)

    def test_create_no_admin(self):
        rule_name = "os_masakari_api:segments:create"
        self.policy.set_rules({rule_name: "project:non_fake"})
        body = {
            "segment": {
                "name": "segment1",
                "service_type": "COMPUTE",
                "recovery_method": "auto",
                "description": "failover_segment for compute"
            }
        }
        exc = self.assertRaises(exception.PolicyNotAuthorized,
                                self.controller.create,
                                self.req, body=body)
        self._check_rule(exc, rule_name)

    def test_show_no_admin(self):
        rule_name = "os_masakari_api:segments:detail"
        self.policy.set_rules({rule_name: "project:non_fake"})
        exc = self.assertRaises(exception.PolicyNotAuthorized,
                                self.controller.show,
                                self.req, uuidsentinel.fake_segment)
        self._check_rule(exc, rule_name)

    def test_update_no_admin(self):
        rule_name = "os_masakari_api:segments:update"
        self.policy.set_rules({rule_name: "project:non_fake"})
        body = {
            "segment": {
                "name": "segment1",
                "service_type": "COMPUTE",
                "recovery_method": "auto",
                "description": "failover_segment for compute"
            }
        }
        exc = self.assertRaises(exception.PolicyNotAuthorized,
                                self.controller.update,
                                self.req, uuidsentinel.fake_segment, body=body)
        self._check_rule(exc, rule_name)

    def test_delete_no_admin(self):
        rule_name = "os_masakari_api:segments:delete"
        self.policy.set_rules({rule_name: "project:non_fake"})
        exc = self.assertRaises(exception.PolicyNotAuthorized,
                                self.controller.delete,
                                self.req, uuidsentinel.fake_segment)
        self._check_rule(exc, rule_name)



# Copyright 2016 NTT DATA
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

import fixtures
from oslo_policy import policy as oslo_policy
from oslo_serialization import jsonutils

import masakari.conf
from masakari.conf import paths
from masakari import policies
import masakari.policy
from masakari.tests.unit import fake_policy

CONF = masakari.conf.CONF


class RealPolicyFixture(fixtures.Fixture):
    """Load the live policy for tests.

    A base policy fixture that starts with the assumption that you'd
    like to load and enforce the shipped default policy in tests.

    Provides interfaces to tinker with both the contents and location
    of the policy file before loading to allow overrides. To do this
    implement ``_prepare_policy`` in the subclass, and adjust the
    ``policy_file`` accordingly.

    """
    def _prepare_policy(self):
        """Allow changing of the policy before we get started"""
        pass

    def setUp(self):
        super(RealPolicyFixture, self).setUp()
        # policy_file can be overridden by subclasses
        self.policy_file = paths.state_path_def('etc/masakari/policy.json')
        self._prepare_policy()
        CONF.set_override('policy_file', self.policy_file, group='oslo_policy')
        masakari.policy.reset()
        masakari.policy.init()
        self.addCleanup(masakari.policy.reset)

    def set_rules(self, rules, overwrite=True):
        policy = masakari.policy._ENFORCER
        policy.set_rules(oslo_policy.Rules.from_dict(rules),
                         overwrite=overwrite)


class PolicyFixture(RealPolicyFixture):
    """Load a fake policy from masakari.tests.unit.fake_policy

    This overrides the policy with a completely fake and synthetic
    policy file.

    """
    def _prepare_policy(self):
        self.policy_dir = self.useFixture(fixtures.TempDir())
        self.policy_file = os.path.join(self.policy_dir.path, 'policy.json')

        # load the fake_policy data and add the missing default rules.
        policy_rules = jsonutils.loads(fake_policy.policy_data)

        with open(self.policy_file, 'w') as f:
            jsonutils.dump(policy_rules, f)
        CONF.set_override('policy_dirs', [], group='oslo_policy')


class RoleBasedPolicyFixture(RealPolicyFixture):
    """Load a modified policy which allows all actions only be a single roll.

    This fixture can be used for testing role based permissions as it
    provides a version of the policy which stomps over all previous
    declaration and makes every action only available to a single
    role.

    """

    def __init__(self, role="admin", *args, **kwargs):
        super(RoleBasedPolicyFixture, self).__init__(*args, **kwargs)
        self.role = role

    def _prepare_policy(self):
        # Convert all actions to require the specified role
        policy = {}
        for rule in policies.list_rules():
            policy[rule.name] = 'role:%s' % self.role

        self.policy_dir = self.useFixture(fixtures.TempDir())
        self.policy_file = os.path.join(self.policy_dir.path, 'policy.json')
        with open(self.policy_file, 'w') as f:
            jsonutils.dump(policy, f)



# Copyright (c) 2016 NTT DATA
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os

from migrate.versioning import api as versioning_api
from migrate.versioning import repository
from oslo_db.sqlalchemy import test_base
from oslo_db.sqlalchemy import test_migrations
from oslo_db.sqlalchemy import utils as oslodbutils
import sqlalchemy
from sqlalchemy.engine import reflection
import sqlalchemy.exc

import masakari.conf
from masakari.db.sqlalchemy import migrate_repo
from masakari.db.sqlalchemy import migration as sa_migration
from masakari.db.sqlalchemy import models
from masakari.tests import fixtures as masakari_fixtures


CONF = masakari.conf.CONF


class MasakariMigrationsCheckers(test_migrations.WalkVersionsMixin):
    """Test sqlalchemy-migrate migrations."""

    TIMEOUT_SCALING_FACTOR = 2

    @property
    def INIT_VERSION(self):
        return sa_migration.INIT_VERSION

    @property
    def REPOSITORY(self):
        return repository.Repository(
            os.path.abspath(os.path.dirname(migrate_repo.__file__)))

    @property
    def migration_api(self):
        return versioning_api

    @property
    def migrate_engine(self):
        return self.engine

    def setUp(self):
        super(MasakariMigrationsCheckers, self).setUp()
        migrate_log = logging.getLogger('migrate')
        old_level = migrate_log.level
        migrate_log.setLevel(logging.WARN)
        self.addCleanup(migrate_log.setLevel, old_level)

        self.useFixture(masakari_fixtures.Timeout(
            os.environ.get('OS_TEST_TIMEOUT', 0),
            self.TIMEOUT_SCALING_FACTOR))

    def assertColumnExists(self, engine, table_name, column):
        self.assertTrue(oslodbutils.column_exists(engine, table_name, column),
                        'Column %s.%s does not exist' % (table_name, column))

    def assertColumnNotExists(self, engine, table_name, column):
        self.assertFalse(oslodbutils.column_exists(engine, table_name, column),
                        'Column %s.%s should not exist' % (table_name, column))

    def assertTableNotExists(self, engine, table):
        self.assertRaises(sqlalchemy.exc.NoSuchTableError,
                          oslodbutils.get_table, engine, table)

    def assertIndexExists(self, engine, table_name, index):
        self.assertTrue(oslodbutils.index_exists(engine, table_name, index),
                        'Index %s on table %s does not exist' %
                        (index, table_name))

    def assertIndexNotExists(self, engine, table_name, index):
        self.assertFalse(oslodbutils.index_exists(engine, table_name, index),
                         'Index %s on table %s should not exist' %
                         (index, table_name))

    def assertIndexMembers(self, engine, table, index, members):
        self.assertIndexExists(engine, table, index)

        t = oslodbutils.get_table(engine, table)
        index_columns = None
        for idx in t.indexes:
            if idx.name == index:
                index_columns = [c.name for c in idx.columns]
                break

        self.assertEqual(members, index_columns)

    def include_object(self, object_, name, type_, reflected, compare_to):
        if type_ == 'table':
            # migrate_version is a sqlalchemy-migrate control table and
            # isn't included in the model. shadow_* are generated from
            # the model and have their own tests to ensure they don't
            # drift.
            if name == 'migrate_version' or name.startswith('shadow_'):
                return False

        return True

    # Implementations for ModelsMigrationsSync
    def db_sync(self, engine):
        sa_migration.db_sync(engine=self.migrate_engine)

    def get_engine(self, context=None):
        return self.migrate_engine

    def get_metadata(self):
        return models.BASE.metadata

    def migrate_up(self, version, with_data=False):
        banned = None

        if with_data:
            check = getattr(self, "_check_%03d" % version, None)
            self.assertIsNotNone(check, ('DB Migration %i does not have a '
                                         'test. Please add one!') % version)

        with masakari_fixtures.BannedDBSchemaOperations(banned):
            super(MasakariMigrationsCheckers, self).migrate_up(version,
                                                               with_data)

    def test_walk_versions(self):
        self.walk_versions(snake_walk=False, downgrade=False)

    def _check_001(self, engine, data):
        self.assertColumnExists(engine, 'failover_segments', 'uuid')
        self.assertColumnExists(engine, 'failover_segments', 'name')
        self.assertColumnExists(engine, 'failover_segments', 'service_type')
        self.assertColumnExists(engine, 'failover_segments', 'description')
        self.assertColumnExists(engine, 'failover_segments',
                                'recovery_method')
        self.assertIndexMembers(engine, 'failover_segments',
                                'segments_service_type_idx', ['service_type'])

    def _check_002(self, engine, data):
        self.assertColumnExists(engine, 'hosts', 'uuid')
        self.assertColumnExists(engine, 'hosts', 'name')
        self.assertColumnExists(engine, 'hosts', 'reserved')
        self.assertColumnExists(engine, 'hosts', 'type')
        self.assertColumnExists(engine, 'hosts', 'control_attributes')
        self.assertColumnExists(engine, 'hosts', 'failover_segment_id')
        self.assertColumnExists(engine, 'hosts', 'on_maintenance')
        self.assertColumnExists(engine, 'hosts', 'type')
        self.assertIndexMembers(engine, 'hosts', 'hosts_type_idx', ['type'])

    def _check_003(self, engine, data):
        inspector = reflection.Inspector.from_engine(engine)
        constraints = inspector.get_unique_constraints('hosts')
        constraint_names = [constraint['name'] for constraint in constraints]
        self.assertIn('uniq_host0name0deleted',
                      constraint_names)

    def _check_004(self, engine, data):
        self.assertColumnExists(engine, 'notifications', 'notification_uuid')
        self.assertColumnExists(engine, 'notifications', 'generated_time')
        self.assertColumnExists(engine, 'notifications', 'source_host_uuid')
        self.assertColumnExists(engine, 'notifications', 'type')
        self.assertColumnExists(engine, 'notifications', 'payload')
        self.assertColumnExists(engine, 'notifications', 'status')

    def _check_005(self, engine, data):
        failover_segments = oslodbutils.get_table(engine, 'failover_segments')
        hosts = oslodbutils.get_table(engine, 'hosts')

        for table in [failover_segments, hosts]:
            self.assertTrue(table.c.created_at.nullable)

    def _check_006(self, engine, data):
        self.assertColumnExists(engine, 'logbooks', 'created_at')
        self.assertColumnExists(engine, 'logbooks', 'updated_at')
        self.assertColumnExists(engine, 'logbooks', 'meta')
        self.assertColumnExists(engine, 'logbooks', 'name')
        self.assertColumnExists(engine, 'logbooks', 'uuid')

        self.assertColumnExists(engine, 'flowdetails', 'created_at')
        self.assertColumnExists(engine, 'flowdetails', 'updated_at')
        self.assertColumnExists(engine, 'flowdetails', 'parent_uuid')
        self.assertColumnExists(engine, 'flowdetails', 'meta')
        self.assertColumnExists(engine, 'flowdetails', 'name')
        self.assertColumnExists(engine, 'flowdetails', 'state')
        self.assertColumnExists(engine, 'flowdetails', 'uuid')

        self.assertColumnExists(engine, 'atomdetails', 'created_at')
        self.assertColumnExists(engine, 'atomdetails', 'updated_at')
        self.assertColumnExists(engine, 'atomdetails', 'parent_uuid')
        self.assertColumnExists(engine, 'atomdetails', 'meta')
        self.assertColumnExists(engine, 'atomdetails', 'name')
        self.assertColumnExists(engine, 'atomdetails', 'results')
        self.assertColumnExists(engine, 'atomdetails', 'version')
        self.assertColumnExists(engine, 'atomdetails', 'state')
        self.assertColumnExists(engine, 'atomdetails', 'uuid')
        self.assertColumnExists(engine, 'atomdetails', 'failure')
        self.assertColumnExists(engine, 'atomdetails', 'atom_type')
        self.assertColumnExists(engine, 'atomdetails', 'intention')
        self.assertColumnExists(engine, 'atomdetails', 'revert_results')
        self.assertColumnExists(engine, 'atomdetails', 'revert_failure')


class TestMasakariMigrationsSQLite(MasakariMigrationsCheckers,
                                   test_base.DbTestCase):

    def _check_006(self, engine, data):
        # NOTE(ShilpaSD): DB script '006_add_persistence_tables.py' adds db
        # tables required for taskflow which doesn't support Sqlite using
        # alembic migration.
        pass

    pass


class TestMasakariMigrationsMySQL(MasakariMigrationsCheckers,
                              test_base.MySQLOpportunisticTestCase):
    def test_innodb_tables(self):
        sa_migration.db_sync(engine=self.migrate_engine)

        total = self.migrate_engine.execute(
            "SELECT count(*) "
            "FROM information_schema.TABLES "
            "WHERE TABLE_SCHEMA = '%(database)s'" %
            {'database': self.migrate_engine.url.database})
        self.assertGreater(total.scalar(), 0, "No tables found. Wrong schema?")

        noninnodb = self.migrate_engine.execute(
            "SELECT count(*) "
            "FROM information_schema.TABLES "
            "WHERE TABLE_SCHEMA='%(database)s' "
            "AND ENGINE != 'InnoDB' "
            "AND TABLE_NAME != 'migrate_version'" %
            {'database': self.migrate_engine.url.database})
        count = noninnodb.scalar()
        self.assertEqual(count, 0, "%d non InnoDB tables created" % count)


class TestMasakariMigrationsPostgreSQL(
    MasakariMigrationsCheckers,
    test_base.PostgreSQLOpportunisticTestCase):
    pass



# Copyright (C) 2018 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_policy import policy

from masakari.policies import base


NOTIFICATIONS = 'os_masakari_api:notifications:%s'

rules = [
    policy.DocumentedRuleDefault(
        name=NOTIFICATIONS % 'index',
        check_str=base.RULE_ADMIN_API,
        description="Lists IDs, notification types, host_name, generated_time,"
                    " payload and status for all notifications.",
        operations=[
            {
                'method': 'GET',
                'path': '/notifications'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=NOTIFICATIONS % 'detail',
        check_str=base.RULE_ADMIN_API,
        description="Shows details for a notification.",
        operations=[
            {
                'method': 'GET',
                'path': '/notifications/{notification_id}'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=NOTIFICATIONS % 'create',
        check_str=base.RULE_ADMIN_API,
        description="Creates a notiification.",
        operations=[
            {
                'method': 'POST',
                'path': '/notifications'
            }
        ]),
    policy.RuleDefault(
        name=NOTIFICATIONS % 'discoverable',
        check_str=base.RULE_ADMIN_API,
        description="Notification API extensions to change the API.",
        ),
]


def list_rules():
    return rules



# Copyright 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Policy Engine For Masakari."""

import copy
import logging
import re
import sys

from oslo_config import cfg
from oslo_policy import policy
from oslo_utils import excutils

from masakari import exception
# from masakari.i18n import _LE, _LW
from masakari import policies


CONF = cfg.CONF
LOG = logging.getLogger(__name__)
_ENFORCER = None
# saved_file_rules and used to compare with new rules to determine the
# rules whether were updated.
saved_file_rules = []
KEY_EXPR = re.compile(r'%\((\w+)\)s')


def reset():
    global _ENFORCER
    if _ENFORCER:
        _ENFORCER.clear()
        _ENFORCER = None


def init(policy_file=None, rules=None, default_rule=None, use_conf=True):
    """Init an Enforcer class.

       :param policy_file: Custom policy file to use, if none is specified,
                           `CONF.policy_file` will be used.
       :param rules: Default dictionary / Rules to use. It will be
                     considered just in the first instantiation.
       :param default_rule: Default rule to use, CONF.default_rule will
                            be used if none is specified.
       :param use_conf: Whether to load rules from config file.
    """
    global _ENFORCER
    global saved_file_rules
    if not _ENFORCER:
        _ENFORCER = policy.Enforcer(CONF,
                                    policy_file=policy_file,
                                    rules=rules,
                                    default_rule=default_rule,
                                    use_conf=use_conf)
        register_rules(_ENFORCER)
        _ENFORCER.load_rules()

    # Only the rules which are loaded from file may be changed.
    current_file_rules = _ENFORCER.file_rules
    current_file_rules = _serialize_rules(current_file_rules)

    # Checks whether the rules are updated in the runtime
    if saved_file_rules != current_file_rules:
        _warning_for_deprecated_user_based_rules(current_file_rules)
        saved_file_rules = copy.deepcopy(current_file_rules)


def _serialize_rules(rules):
    """Serialize all the Rule object as string which is used to compare the
    rules list.
    """
    result = [(rule_name, str(rule))
              for rule_name, rule in rules.items()]
    return sorted(result, key=lambda rule: rule[0])


def _warning_for_deprecated_user_based_rules(rules):
    """Warning user based policy enforcement used in the rule but the rule
    doesn't support it.
    """
    for rule in rules:
        if 'user_id' in KEY_EXPR.findall(rule[1]):
            LOG.debug(("The user_id attribute isn't supported in the rule%s'. "
                       "All the user_id based policy enforcement will be "
                       "removed in the future."), rule[0])


def set_rules(rules, overwrite=True, use_conf=False):
    """Set rules based on the provided dict of rules.

       :param rules: New rules to use. It should be an instance of dict.
       :param overwrite: Whether to overwrite current rules or update them
                         with the new rules.
       :param use_conf: Whether to reload rules from config file.
    """

    init(use_conf=False)
    _ENFORCER.set_rules(rules, overwrite, use_conf)


def authorize(context, action, target, do_raise=True, exc=None):
    """Verifies that the action is valid on the target in this context.

       :param context: masakari context
       :param action: string representing the action to be checked
           this should be colon separated for clarity.
           i.e. ``os_masakari_api:segments``,
           ``os_masakari_api:os-hosts``,
           ``os_masakari_api:notifications``,
           ``os_masakari_api:extensions``
       :param target: dictionary representing the object of the action
           for object creation this should be a dictionary representing the
           location of the object e.g. ``{'project_id': context.project_id}``
       :param do_raise: if True (the default), raises PolicyNotAuthorized;
           if False, returns False
       :param exc: Class of the exception to raise if the check fails.
                   Any remaining arguments passed to :meth:`authorize` (both
                   positional and keyword arguments) will be passed to
                   the exception class. If not specified,
                   :class:`PolicyNotAuthorized` will be used.

       :raises masakari.exception.PolicyNotAuthorized: if verification fails
           and do_raise is True. Or if 'exc' is specified it will raise an
           exception of that type.

       :return: returns a non-False value (not necessarily "True") if
           authorized, and the exact value False if not authorized and
           do_raise is False.
    """
    init()
    credentials = context.to_policy_values()
    if not exc:
        exc = exception.PolicyNotAuthorized
    try:
        result = _ENFORCER.authorize(action, target, credentials,
                                     do_raise=do_raise, exc=exc, action=action)
    except policy.PolicyNotRegistered:
        with excutils.save_and_reraise_exception():
            LOG.debug('Policy not registered')
    except Exception:
        with excutils.save_and_reraise_exception():
            LOG.debug('Policy check for %(action)s failed with credentials '
                      '%(credentials)s',
                      {'action': action, 'credentials': credentials})
    return result


def check_is_admin(context):
    """Whether or not roles contains 'admin' role according to policy setting.

    """

    init()
    # the target is user-self
    credentials = context.to_policy_values()
    target = credentials
    return _ENFORCER.authorize('context_is_admin', target, credentials)


@policy.register('is_admin')
class IsAdminCheck(policy.Check):
    """An explicit check for is_admin."""

    def __init__(self, kind, match):
        """Initialize the check."""

        self.expected = (match.lower() == 'true')

        super(IsAdminCheck, self).__init__(kind, str(self.expected))

    def __call__(self, target, creds, enforcer):
        """Determine whether is_admin matches the requested value."""

        return creds['is_admin'] == self.expected


def get_rules():
    if _ENFORCER:
        return _ENFORCER.rules


def register_rules(enforcer):
    enforcer.register_defaults(policies.list_rules())


def get_enforcer():
    # This method is for use by oslopolicy CLI scripts. Those scripts need the
    # 'output-file' and 'namespace' options, but having those in sys.argv means
    # loading the Masakari config options will fail as those are not expected
    # to be present. So we pass in an arg list with those stripped out.
    conf_args = []
    # Start at 1 because cfg.CONF expects the equivalent of sys.argv[1:]
    i = 1
    while i < len(sys.argv):
        if sys.argv[i].strip('-') in ['namespace', 'output-file']:
            i += 2
            continue
        conf_args.append(sys.argv[i])
        i += 1

    cfg.CONF(conf_args, project='masakari')
    init()
    return _ENFORCER



#    Copyright (c) 2016 NTT DATA
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit Tests for masakari.engine.rpcapi
"""

import copy

import mock

from masakari import context
from masakari.engine import rpcapi as engine_rpcapi
from masakari import objects
from masakari import test
from masakari.tests.unit.engine import fake_engine


class EngineRpcAPITestCase(test.TestCase):
    def setUp(self):
        super(EngineRpcAPITestCase, self).setUp()
        self.context = context.RequestContext()
        self.fake_notification_obj = fake_engine.fake_notification_obj(
            self.context)

    def _test_engine_api(self, method, rpc_method, server=None, fanout=False,
                         **kwargs):
        rpcapi = engine_rpcapi.EngineAPI()
        expected_retval = 'foo' if rpc_method == 'call' else None

        target = {
            "server": server,
            "fanout": fanout,
            "version": kwargs.pop('version', rpcapi.RPC_API_VERSION)
        }

        expected_msg = copy.deepcopy(kwargs)

        self.fake_args = None
        self.fake_kwargs = None

        def _fake_prepare_method(*args, **kwds):
            for kwd in kwds:
                self.assertEqual(target[kwd], kwds[kwd])
            return rpcapi.client

        def _fake_rpc_method(*args, **kwargs):
            self.fake_args = args
            self.fake_kwargs = kwargs
            if expected_retval:
                return expected_retval

        with mock.patch.object(rpcapi.client, "prepare") as mock_prepared:
            mock_prepared.side_effect = _fake_prepare_method

            with mock.patch.object(rpcapi.client, rpc_method) as mock_method:
                mock_method.side_effect = _fake_rpc_method
                retval = getattr(rpcapi, method)(self.context, **kwargs)
                self.assertEqual(expected_retval, retval)
                expected_args = [self.context, method, expected_msg]
                for arg, expected_arg in zip(self.fake_args, expected_args):
                    self.assertEqual(expected_arg, arg)

                for kwarg, value in self.fake_kwargs.items():
                    if isinstance(value, objects.Notification):
                        expected_back = expected_msg[kwarg].obj_to_primitive()
                        backup = value.obj_to_primitive()
                        self.assertEqual(expected_back, backup)
                    else:
                        self.assertEqual(expected_msg[kwarg], value)

    @mock.patch("masakari.rpc.get_client")
    def test_process_notification(self, mock_get_client):
        self._test_engine_api('process_notification',
                              rpc_method='cast',
                              notification=self.fake_notification_obj,
                              version='1.0')

    @mock.patch("masakari.rpc.get_client")
    def test_get_notification_recovery_workflow_details(self,
                                                        mock_get_client):
        self._test_engine_api('get_notification_recovery_workflow_details',
                              rpc_method='call',
                              notification=self.fake_notification_obj,
                              version='1.1')



# Copyright 2016 NTT Data.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import timeutils
from six.moves import http_client as http
from webob import exc

from masakari.api import api_version_request
from masakari.api.openstack import common
from masakari.api.openstack import extensions
from masakari.api.openstack.ha.schemas import notifications as schema
from masakari.api.openstack.ha.schemas import payload as payload_schema
from masakari.api.openstack import wsgi
from masakari.api import validation
from masakari import exception
from masakari.ha import api as notification_api
from masakari.i18n import _
from masakari.objects import fields
from masakari.policies import notifications as notifications_policies

ALIAS = 'notifications'


class NotificationsController(wsgi.Controller):
    """Notifications controller for the OpenStack API."""

    def __init__(self):
        self.api = notification_api.NotificationAPI()

    @validation.schema(payload_schema.create_process_payload)
    def _validate_process_payload(self, req, body):
        pass

    @validation.schema(payload_schema.create_vm_payload)
    def _validate_vm_payload(self, req, body):
        pass

    @validation.schema(payload_schema.create_compute_host_payload)
    def _validate_comp_host_payload(self, req, body):
        pass

    @wsgi.response(http.ACCEPTED)
    @extensions.expected_errors((http.BAD_REQUEST, http.FORBIDDEN,
                                 http.CONFLICT))
    @validation.schema(schema.create)
    def create(self, req, body):
        """Creates a new notification."""
        context = req.environ['masakari.context']
        context.can(notifications_policies.NOTIFICATIONS % 'create')

        notification_data = body['notification']
        if notification_data['type'] == fields.NotificationType.PROCESS:
            self._validate_process_payload(req,
                                           body=notification_data['payload'])

        if notification_data['type'] == fields.NotificationType.VM:
            self._validate_vm_payload(req, body=notification_data['payload'])

        if notification_data['type'] == fields.NotificationType.COMPUTE_HOST:
            self._validate_comp_host_payload(req,
                                             body=notification_data['payload'])

        try:
            notification = self.api.create_notification(
                context, notification_data)
        except exception.HostNotFoundByName as err:
            raise exc.HTTPBadRequest(explanation=err.format_message())
        except (exception.DuplicateNotification,
                exception.HostOnMaintenanceError) as err:
            raise exc.HTTPConflict(explanation=err.format_message())

        return {'notification': notification}

    @extensions.expected_errors((http.BAD_REQUEST, http.FORBIDDEN))
    def index(self, req):
        """Returns a summary list of notifications."""
        context = req.environ['masakari.context']
        context.can(notifications_policies.NOTIFICATIONS % 'index')
        try:
            limit, marker = common.get_limit_and_marker(req)
            sort_keys, sort_dirs = common.get_sort_params(req.params)

            filters = {}
            if 'status' in req.params:
                filters['status'] = req.params['status']
            if 'source_host_uuid' in req.params:
                filters['source_host_uuid'] = req.params['source_host_uuid']
            if 'type' in req.params:
                filters['type'] = req.params['type']
            if 'generated-since' in req.params:
                try:
                    parsed = timeutils.parse_isotime(
                        req.params['generated-since'])
                except ValueError:
                    msg = _('Invalid generated-since value')
                    raise exc.HTTPBadRequest(explanation=msg)
                filters['generated-since'] = parsed

            notifications = self.api.get_all(context, filters, sort_keys,
                                             sort_dirs, limit, marker)
        except exception.MarkerNotFound as err:
            raise exc.HTTPBadRequest(explanation=err.format_message())
        except exception.Invalid as err:
            raise exc.HTTPBadRequest(explanation=err.format_message())

        return {'notifications': notifications}

    @extensions.expected_errors((http.FORBIDDEN, http.NOT_FOUND))
    def show(self, req, id):
        """Return data about the given notification id."""
        context = req.environ['masakari.context']
        context.can(notifications_policies.NOTIFICATIONS % 'detail')

        try:
            if api_version_request.is_supported(req, min_version='1.1'):
                notification = (
                    self.api.get_notification_recovery_workflow_details(
                        context, id))
            else:
                notification = self.api.get_notification(context, id)
        except exception.NotificationNotFound as err:
            raise exc.HTTPNotFound(explanation=err.format_message())
        return {'notification': notification}


class Notifications(extensions.V1APIExtensionBase):
    """Notifications support."""

    name = "Notifications"
    alias = ALIAS
    version = 1

    def get_resources(self):
        member_actions = {'action': 'POST'}

        resources = [
            extensions.ResourceExtension(ALIAS,
                                         NotificationsController(),
                                         member_name='notification',
                                         member_actions=member_actions)
            ]
        return resources

    def get_controller_extensions(self):
        return []



# Copyright 2016 NTT DATA
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Matcher classes to be used inside of the testtools assertThat framework."""

import six
import testtools.matchers


class EncodedByUTF8(object):
    def match(self, obj):
        if isinstance(obj, six.binary_type):
            if hasattr(obj, "decode"):
                try:
                    obj.decode("utf-8")
                except UnicodeDecodeError:
                    return testtools.matchers.Mismatch(
                        "%s is not encoded in UTF-8." % obj)
        elif isinstance(obj, six.text_type):
            try:
                obj.encode("utf-8", "strict")
            except UnicodeDecodeError:
                return testtools.matchers.Mismatch("%s cannot be "
                                                   "encoded in UTF-8." % obj)
        else:
            reason = ("Type of '%(obj)s' is '%(obj_type)s', "
                      "should be '%(correct_type)s'."
                      % {
                          "obj": obj,
                          "obj_type": type(obj).__name__,
                          "correct_type": six.binary_type.__name__
                      })
            return testtools.matchers.Mismatch(reason)



# Copyright 2016 NTT DATA
# All Rights Reserved.

#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
import six
from six.moves import builtins

from masakari import test
from masakari import version


class VersionTestCase(test.NoDBTestCase):
    """Test cases for Versions code."""

    def test_version_string_with_package_is_good(self):
        """Ensure uninstalled code get version string."""

        self.stub_out('masakari.version.version_info.version_string',
                lambda: '5.5.5.5')
        self.stub_out('masakari.version.MASAKARI_PACKAGE', 'g9ec3421')
        self.assertEqual("5.5.5.5-g9ec3421",
                         version.version_string_with_package())

    def test_release_file(self):
        version.loaded = False
        real_open = builtins.open
        real_find_file = cfg.CONF.find_file

        def fake_find_file(self, name):
            if name == "release":
                return "/etc/masakari/release"
            return real_find_file(self, name)

        def fake_open(path, *args, **kwargs):
            if path == "/etc/masakari/release":
                data = """[Masakari]
vendor = ACME Corporation
product = ACME Masakari
package = 1337"""
                return six.StringIO(data)

            return real_open(path, *args, **kwargs)

        self.stub_out('six.moves.builtins.open', fake_open)
        self.stub_out('oslo_config.cfg.ConfigOpts.find_file', fake_find_file)

        self.assertEqual(version.vendor_string(), "ACME Corporation")
        self.assertEqual(version.product_string(), "ACME Masakari")
        self.assertEqual(version.package_string(), "1337")



# Copyright (c) 2017 NTT Data
# All Rights Reserved.
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# masakari documentation build configuration file.
#
# This file is execfile()d with the current directory set to
# its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import subprocess
import sys
import warnings

import openstackdocstheme

from masakari.version import version_info


extensions = [
    'os_api_ref',
]


html_theme = 'openstackdocs'
html_theme_path = [openstackdocstheme.get_html_theme_path()]
html_theme_options = {
    "sidebar_mode": "toc",
}
html_context = {'bug_project': 'masakari', 'bug_tag': 'api-ref'}

# If extensions (or modules to document with autodoc) are in another
# directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown
# here.
sys.path.insert(0, os.path.abspath('../../'))
sys.path.insert(0, os.path.abspath('../'))
sys.path.insert(0, os.path.abspath('./'))

# -- General configuration
# ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#
# source_encoding = 'utf-8'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'Masakari API Reference'
copyright = u'2017-present, OpenStack Foundation'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tags.
release = version_info.release_string()
# The short X.Y version.
version = version_info.version_string()

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# The reST default role (used for this markup: `text`) to use
# for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = False

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# -- Options for man page output
# ----------------------------------------------

# Grouping the document tree for man pages.
# List of tuples 'sourcefile', 'target', u'title', u'Authors name', 'manual'


# -- Options for HTML output
# --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
# html_theme_path = ["."]
# html_theme = '_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'
git_cmd = ["git", "log", "--pretty=format:'%ad, commit %h'", "--date=local",
           "-n1"]
try:
    html_last_updated_fmt = subprocess.Popen(
        git_cmd, stdout=subprocess.PIPE).communicate()[0].decode()
except Exception:
    warnings.warn('Cannot get last updated time from git repository. '
                  'Not setting "html_last_updated_fmt".')

    # If true, SmartyPants will be used to convert quotes and dashes to
    # typographically correct entities.
    # html_use_smartypants = True

    # Custom sidebar templates, maps document names to template names.
    # html_sidebars = {}

    # Additional templates that should be rendered to pages, maps page
    # names to
    # template names.
    # html_additional_pages = {}

    # If false, no module index is generated.
    # html_use_modindex = True

    # If false, no index is generated.
    # html_use_index = True

    # If true, the index is split into individual pages for each letter.
    # html_split_index = False

    # If true, links to the reST sources are added to the pages.
    # html_show_sourcelink = True

    # If true, an OpenSearch description file will be output, and all
    # pages will
    # contain a <link> tag referring to it.  The value of this option must
    # be the
    # base URL from which the finished HTML is served.
    # html_use_opensearch = ''

    # If nonempty, this is the file name suffix for HTML files (e.g.
    # ".xhtml").
    # html_file_suffix = ''

    # Output file base name for HTML help builder.
    htmlhelp_basename = 'masakaridoc'

    # -- Options for LaTeX output
    # -------------------------------------------------

    # The paper size ('letter' or 'a4').
    # latex_paper_size = 'letter'

    # The font size ('10pt', '11pt' or '12pt').
    # latex_font_size = '10pt'

    # Grouping the document tree into LaTeX files. List of tuples
    # (source start file, target name, title, author, documentclass
    # [howto/manual]).
    latex_documents = [('index', 'Masakari.tex',
                        u'OpenStack Masakari API Documentation',
                        u'OpenStack Foundation', 'manual'),
                       ]

    # The name of an image file (relative to this directory) to place at
    # the top of
    # the title page.
    # latex_logo = None

    # For "manual" documents, if this is true, then toplevel headings are
    # parts,
    # not chapters.
    # latex_use_parts = False

    # Additional stuff for the LaTeX preamble.
    # latex_preamble = ''

    # Documents to append as an appendix to all manuals.
    # latex_appendices = []

    # If false, no module index is generated.
    # latex_use_modindex = True



# Copyright 2019 NTT Data.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import masakari.conf
from masakari.engine import driver

CONF = masakari.conf.CONF
NOTIFICATION_DRIVER = CONF.notification_driver
PERSISTENCE_BACKEND = CONF.taskflow.connection


def upgrade(migrate_engine):
    """Upgrade the engine with persistence tables. """

    # Get the taskflow driver configured, default is 'taskflow_driver',
    # to load persistence tables to store progress details.
    taskflow_driver = driver.load_masakari_driver(NOTIFICATION_DRIVER)

    if PERSISTENCE_BACKEND:
        taskflow_driver.upgrade_backend(PERSISTENCE_BACKEND)



#    Copyright 2011 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import errno
import platform
import socket
import sys

import masakari.conf
import masakari.context
import masakari.utils

CONF = masakari.conf.CONF


def get_test_admin_context():
    return masakari.context.get_admin_context()


def is_linux():
    return platform.system() == 'Linux'


def coreutils_readlink_available():
    _out, err = masakari.utils.trycmd('readlink', '-nm', '/')
    return err == ''


def is_ipv6_supported():
    has_ipv6_support = socket.has_ipv6
    try:
        s = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
        s.close()
    except socket.error as e:
        if e.errno == errno.EAFNOSUPPORT:
            has_ipv6_support = False
        else:
            raise

    # check if there is at least one interface with ipv6
    if has_ipv6_support and sys.platform.startswith('linux'):
        try:
            with open('/proc/net/if_inet6') as f:
                if not f.read():
                    has_ipv6_support = False
        except IOError:
            has_ipv6_support = False

    return has_ipv6_support


def get_api_version(request):
    if request.path[2:3].isdigit():
        return int(request.path[2:3])


def _compare_args(args1, args2, cmp):
    return all(cmp(*pair) for pair in zip(args1, args2))


def _compare_kwargs(kwargs1, kwargs2, cmp):
    return all(cmp(kwargs1[k], kwargs2[k])
               for k in set(list(kwargs1.keys()) + list(kwargs2.keys())))



# All Rights Reserved.
# Copyright 2016 NTT DATA
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import mock
from oslo_utils import importutils
from oslo_utils import timeutils

from masakari.compute import nova
import masakari.conf
from masakari import context
from masakari.engine import manager
from masakari.engine import utils as engine_utils
from masakari import exception
from masakari.objects import fields
from masakari.objects import host as host_obj
from masakari.objects import notification as notification_obj
from masakari import rpc
from masakari import test
from masakari.tests.unit import fakes
from masakari.tests import uuidsentinel

CONF = masakari.conf.CONF

NOW = timeutils.utcnow().replace(microsecond=0)


def _get_vm_type_notification(status="new"):
    return fakes.create_fake_notification(
        type="VM", id=1, payload={
            'event': 'LIFECYCLE', 'instance_uuid': uuidsentinel.fake_ins,
            'vir_domain_event': 'STOPPED_FAILED'
        },
        source_host_uuid=uuidsentinel.fake_host,
        generated_time=NOW, status=status,
        notification_uuid=uuidsentinel.fake_notification)


@mock.patch.object(notification_obj.Notification, "get_by_uuid")
class EngineManagerUnitTestCase(test.NoDBTestCase):
    def setUp(self):
        super(EngineManagerUnitTestCase, self).setUp()
        rpc.init(CONF)
        self.engine = importutils.import_object(CONF.engine_manager)
        self.context = context.RequestContext()

    def _fake_notification_workflow(self, exc=None):
        if exc:
            return exc
        # else the workflow executed successfully

    def _get_process_type_notification(self):
        return fakes.create_fake_notification(
            type="PROCESS", id=1, payload={
                'event': 'stopped', 'process_name': 'fake_service'
            },
            source_host_uuid=uuidsentinel.fake_host,
            generated_time=NOW, status="new",
            notification_uuid=uuidsentinel.fake_notification)

    def _get_compute_host_type_notification(self):
        return fakes.create_fake_notification(
            type="COMPUTE_HOST", id=1, payload={
                'event': 'stopped', 'host_status': 'NORMAL',
                'cluster_status': 'ONLINE'
            },
            source_host_uuid=uuidsentinel.fake_host,
            generated_time=NOW, status="new",
            notification_uuid=uuidsentinel.fake_notification)

    @mock.patch("masakari.engine.drivers.taskflow."
                "TaskFlowDriver.execute_instance_failure")
    @mock.patch.object(notification_obj.Notification, "save")
    @mock.patch.object(engine_utils, 'notify_about_notification_update')
    def test_process_notification_type_vm_success(self,
                            mock_notify_about_notification_update, mock_save,
                            mock_instance_failure, mock_notification_get):
        mock_instance_failure.side_effect = self._fake_notification_workflow()
        notification = _get_vm_type_notification()
        mock_notification_get.return_value = notification
        self.engine.process_notification(self.context,
                                         notification=notification)
        self.assertEqual("finished", notification.status)
        mock_instance_failure.assert_called_once_with(
            self.context, notification.payload.get('instance_uuid'),
            notification.notification_uuid)
        action = fields.EventNotificationAction.NOTIFICATION_PROCESS
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, notification, action=action,
                      phase=phase_start),
            mock.call(self.context, notification, action=action,
                      phase=phase_end)]
        mock_notify_about_notification_update.assert_has_calls(notify_calls)

    @mock.patch("masakari.engine.drivers.taskflow."
                "TaskFlowDriver.execute_instance_failure")
    @mock.patch.object(notification_obj.Notification, "save")
    @mock.patch.object(engine_utils, 'notify_about_notification_update')
    @mock.patch('traceback.format_exc')
    def test_process_notification_type_vm_error(self, mock_format,
                            mock_notify_about_notification_update, mock_save,
                            mock_instance_failure, mock_notification_get):
        mock_format.return_value = mock.ANY
        mock_instance_failure.side_effect = self._fake_notification_workflow(
            exc=exception.InstanceRecoveryFailureException)
        notification = _get_vm_type_notification()
        mock_notification_get.return_value = notification
        self.engine.process_notification(self.context,
                                         notification=notification)
        self.assertEqual("error", notification.status)
        mock_instance_failure.assert_called_once_with(
            self.context, notification.payload.get('instance_uuid'),
            notification.notification_uuid)
        e = exception.InstanceRecoveryFailureException('Failed to execute '
                                                'instance recovery workflow.')
        action = fields.EventNotificationAction.NOTIFICATION_PROCESS
        phase_start = fields.EventNotificationPhase.START
        phase_error = fields.EventNotificationPhase.ERROR
        notify_calls = [
            mock.call(self.context, notification, action=action,
                      phase=phase_start),
            mock.call(self.context, notification, action=action,
                      phase=phase_error,
                      exception=str(e),
                      tb=mock.ANY)]
        mock_notify_about_notification_update.assert_has_calls(notify_calls)

    @mock.patch.object(notification_obj.Notification, "save")
    def test_process_notification_type_vm_error_event_unmatched(
            self, mock_save, mock_notification_get):
        notification = fakes.create_fake_notification(
            type="VM", id=1, payload={
                'event': 'fake_event', 'instance_uuid': uuidsentinel.fake_ins,
                'vir_domain_event': 'fake_vir_domain_event'
            },
            source_host_uuid=uuidsentinel.fake_host,
            generated_time=NOW, status="new",
            notification_uuid=uuidsentinel.fake_notification)

        mock_notification_get.return_value = notification
        self.engine.process_notification(self.context,
                                         notification=notification)
        self.assertEqual("ignored", notification.status)

    @mock.patch("masakari.engine.drivers.taskflow."
                "TaskFlowDriver.execute_instance_failure")
    @mock.patch.object(notification_obj.Notification, "save")
    @mock.patch.object(engine_utils, 'notify_about_notification_update')
    @mock.patch('traceback.format_exc')
    def test_process_notification_type_vm_skip_recovery(
            self, mock_format, mock_notify_about_notification_update,
            mock_save, mock_instance_failure, mock_notification_get):
        mock_format.return_value = mock.ANY
        notification = _get_vm_type_notification()
        mock_notification_get.return_value = notification
        mock_instance_failure.side_effect = self._fake_notification_workflow(
            exc=exception.SkipInstanceRecoveryException)
        self.engine.process_notification(self.context,
                                         notification=notification)
        self.assertEqual("finished", notification.status)
        mock_instance_failure.assert_called_once_with(
            self.context, notification.payload.get('instance_uuid'),
            notification.notification_uuid)
        action = fields.EventNotificationAction.NOTIFICATION_PROCESS
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, notification, action=action,
                      phase=phase_start),
            mock.call(self.context, notification, action=action,
                      phase=phase_end)]
        mock_notify_about_notification_update.assert_has_calls(notify_calls)

    @mock.patch.object(host_obj.Host, "get_by_uuid")
    @mock.patch.object(host_obj.Host, "save")
    @mock.patch("masakari.engine.drivers.taskflow."
                "TaskFlowDriver.execute_process_failure")
    @mock.patch.object(notification_obj.Notification, "save")
    @mock.patch.object(engine_utils, 'notify_about_notification_update')
    def test_process_notification_type_process_event_stopped(
            self, mock_notify_about_notification_update,
            mock_notification_save, mock_process_failure,
            mock_host_save, mock_host_obj, mock_notification_get):
        notification = self._get_process_type_notification()
        mock_notification_get.return_value = notification
        mock_process_failure.side_effect = self._fake_notification_workflow()
        fake_host = fakes.create_fake_host()
        mock_host_obj.return_value = fake_host
        self.engine.process_notification(self.context,
                                         notification=notification)
        self.assertEqual("finished", notification.status)
        mock_host_save.assert_called_once()
        mock_process_failure.assert_called_once_with(
            self.context, notification.payload.get('process_name'),
            fake_host.name,
            notification.notification_uuid)
        action = fields.EventNotificationAction.NOTIFICATION_PROCESS
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, notification, action=action,
                      phase=phase_start),
            mock.call(self.context, notification, action=action,
                      phase=phase_end)]
        mock_notify_about_notification_update.assert_has_calls(notify_calls)

    @mock.patch.object(host_obj.Host, "get_by_uuid")
    @mock.patch.object(host_obj.Host, "save")
    @mock.patch("masakari.engine.drivers.taskflow."
                "TaskFlowDriver.execute_process_failure")
    @mock.patch.object(notification_obj.Notification, "save")
    @mock.patch.object(engine_utils, 'notify_about_notification_update')
    @mock.patch('traceback.format_exc')
    def test_process_notification_type_process_skip_recovery(
            self, mock_format, mock_notify_about_notification_update,
            mock_notification_save, mock_process_failure,
            mock_host_save, mock_host_obj, mock_notification_get):
        mock_format.return_value = mock.ANY
        notification = self._get_process_type_notification()
        mock_notification_get.return_value = notification
        fake_host = fakes.create_fake_host()
        mock_host_obj.return_value = fake_host
        mock_process_failure.side_effect = self._fake_notification_workflow(
            exc=exception.SkipProcessRecoveryException)
        self.engine.process_notification(self.context,
                                         notification=notification)
        self.assertEqual("finished", notification.status)
        mock_host_save.assert_called_once()
        action = fields.EventNotificationAction.NOTIFICATION_PROCESS
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, notification, action=action,
                      phase=phase_start),
            mock.call(self.context, notification, action=action,
                      phase=phase_end)]
        mock_notify_about_notification_update.assert_has_calls(notify_calls)

    @mock.patch.object(host_obj.Host, "get_by_uuid")
    @mock.patch.object(host_obj.Host, "save")
    @mock.patch("masakari.engine.drivers.taskflow."
                "TaskFlowDriver.execute_process_failure")
    @mock.patch.object(notification_obj.Notification, "save")
    @mock.patch.object(engine_utils, 'notify_about_notification_update')
    @mock.patch('traceback.format_exc')
    def test_process_notification_type_process_recovery_failure(
            self, mock_format, mock_notify_about_notification_update,
            mock_notification_save, mock_process_failure,
            mock_host_save, mock_host_obj, mock_notification_get):
        mock_format.return_value = mock.ANY
        notification = self._get_process_type_notification()
        mock_notification_get.return_value = notification
        fake_host = fakes.create_fake_host()
        mock_host_obj.return_value = fake_host
        mock_process_failure.side_effect = self._fake_notification_workflow(
            exc=exception.ProcessRecoveryFailureException)
        self.engine.process_notification(self.context,
                                         notification=notification)
        self.assertEqual("error", notification.status)
        mock_host_save.assert_called_once()
        e = exception.ProcessRecoveryFailureException('Failed to execute '
                                                'process recovery workflow.')
        action = fields.EventNotificationAction.NOTIFICATION_PROCESS
        phase_start = fields.EventNotificationPhase.START
        phase_error = fields.EventNotificationPhase.ERROR
        notify_calls = [
            mock.call(self.context, notification, action=action,
                      phase=phase_start),
            mock.call(self.context, notification, action=action,
                      phase=phase_error,
                      exception=str(e),
                      tb=mock.ANY)]
        mock_notify_about_notification_update.assert_has_calls(notify_calls)

    @mock.patch.object(host_obj.Host, "get_by_uuid")
    @mock.patch.object(notification_obj.Notification, "save")
    @mock.patch.object(engine_utils, 'notify_about_notification_update')
    @mock.patch("masakari.engine.drivers.taskflow."
                "TaskFlowDriver.execute_process_failure")
    def test_process_notification_type_process_event_started(
            self, mock_process_failure, mock_notify_about_notification_update,
            mock_notification_save, mock_host_obj,
            mock_notification_get):
        notification = self._get_process_type_notification()
        mock_notification_get.return_value = notification
        notification.payload['event'] = 'started'
        fake_host = fakes.create_fake_host()
        mock_host_obj.return_value = fake_host
        self.engine.process_notification(self.context,
                                         notification=notification)
        self.assertEqual("finished", notification.status)
        self.assertFalse(mock_process_failure.called)
        action = fields.EventNotificationAction.NOTIFICATION_PROCESS
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, notification, action=action,
                      phase=phase_start),
            mock.call(self.context, notification, action=action,
                      phase=phase_end)]
        mock_notify_about_notification_update.assert_has_calls(notify_calls)

    @mock.patch.object(notification_obj.Notification, "save")
    @mock.patch.object(engine_utils, 'notify_about_notification_update')
    @mock.patch("masakari.engine.drivers.taskflow."
                "TaskFlowDriver.execute_process_failure")
    def test_process_notification_type_process_event_other(
            self, mock_process_failure, mock_notify_about_notification_update,
            mock_notification_save, mock_notification_get):
        notification = self._get_process_type_notification()
        mock_notification_get.return_value = notification
        notification.payload['event'] = 'other'
        self.engine.process_notification(self.context,
                                         notification=notification)
        self.assertEqual("ignored", notification.status)
        self.assertFalse(mock_process_failure.called)
        action = fields.EventNotificationAction.NOTIFICATION_PROCESS
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, notification, action=action,
                      phase=phase_start),
            mock.call(self.context, notification, action=action,
                      phase=phase_end)]
        mock_notify_about_notification_update.assert_has_calls(notify_calls)

    @mock.patch.object(host_obj.Host, "get_by_uuid")
    @mock.patch.object(host_obj.Host, "save")
    @mock.patch.object(host_obj.Host, "update")
    @mock.patch.object(host_obj.HostList, "get_all")
    @mock.patch("masakari.engine.drivers.taskflow."
                "TaskFlowDriver.execute_host_failure")
    @mock.patch.object(notification_obj.Notification, "save")
    @mock.patch.object(engine_utils, 'notify_about_notification_update')
    def test_process_notification_type_compute_host_event_stopped(
            self, mock_notify_about_notification_update,
            mock_notification_save, mock_host_failure, mock_get_all,
            mock_host_update, mock_host_save, mock_host_obj,
            mock_notification_get):
        notification = self._get_compute_host_type_notification()
        mock_notification_get.return_value = notification
        mock_host_failure.side_effect = self._fake_notification_workflow()
        fake_host = fakes.create_fake_host()
        mock_get_all.return_value = None
        fake_host.failover_segment = fakes.create_fake_failover_segment()
        mock_host_obj.return_value = fake_host
        self.engine.process_notification(self.context,
                                         notification=notification)

        update_data_by_host_failure = {
            'on_maintenance': True,
        }
        mock_host_update.assert_called_once_with(update_data_by_host_failure)
        mock_host_save.assert_called_once()
        self.assertEqual("finished", notification.status)
        mock_host_failure.assert_called_once_with(
            self.context,
            fake_host.name, fake_host.failover_segment.recovery_method,
            notification.notification_uuid, reserved_host_list=None,
            update_host_method=manager.update_host_method)
        action = fields.EventNotificationAction.NOTIFICATION_PROCESS
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, notification, action=action,
                      phase=phase_start),
            mock.call(self.context, notification, action=action,
                      phase=phase_end)]
        mock_notify_about_notification_update.assert_has_calls(notify_calls)

    @mock.patch.object(host_obj.Host, "get_by_uuid")
    @mock.patch.object(host_obj.Host, "save")
    @mock.patch.object(host_obj.Host, "update")
    @mock.patch.object(host_obj.HostList, "get_all")
    @mock.patch.object(notification_obj.Notification, "save")
    @mock.patch.object(engine_utils, 'notify_about_notification_update')
    @mock.patch('traceback.format_exc')
    def test_process_notification_host_failure_without_reserved_hosts(
            self, mock_format, mock_notify_about_notification_update,
            mock_notification_save, mock_get_all,
            mock_host_update, mock_host_save, mock_host_obj,
            mock_notification_get):
        mock_format.return_value = mock.ANY

        fake_host = fakes.create_fake_host()
        fake_host.failover_segment = fakes.create_fake_failover_segment(
            recovery_method='reserved_host')
        mock_host_obj.return_value = fake_host

        notification = self._get_compute_host_type_notification()
        mock_notification_get.return_value = notification

        self.engine.process_notification(self.context,
                                         notification=notification)

        update_data_by_host_failure = {
            'on_maintenance': True,
        }
        mock_host_update.assert_called_once_with(update_data_by_host_failure)
        mock_host_save.assert_called_once()
        self.assertEqual("error", notification.status)
        action = fields.EventNotificationAction.NOTIFICATION_PROCESS
        phase_start = fields.EventNotificationPhase.START
        phase_error = fields.EventNotificationPhase.ERROR
        e = exception.ReservedHostsUnavailable(
            'No reserved_hosts available for evacuation.')
        notify_calls = [
            mock.call(self.context, notification, action=action,
                      phase=phase_start),
            mock.call(self.context, notification, action=action,
                      phase=phase_error,
                      exception=str(e),
                      tb=mock.ANY)]
        mock_notify_about_notification_update.assert_has_calls(notify_calls)

    @mock.patch.object(host_obj.Host, "get_by_uuid")
    @mock.patch.object(host_obj.Host, "save")
    @mock.patch.object(host_obj.Host, "update")
    @mock.patch.object(host_obj.HostList, "get_all")
    @mock.patch("masakari.engine.drivers.taskflow."
                "TaskFlowDriver.execute_host_failure")
    @mock.patch.object(notification_obj.Notification, "save")
    @mock.patch.object(engine_utils, 'notify_about_notification_update')
    def test_process_notification_host_failure_with_reserved_hosts(
            self, mock_notify_about_notification_update,
            mock_notification_save, mock_host_failure, mock_get_all,
            mock_host_update, mock_host_save, mock_host_obj,
            mock_notification_get):
        fake_host = fakes.create_fake_host()
        fake_host.failover_segment = fakes.create_fake_failover_segment(
            recovery_method='reserved_host')
        reserved_host_object_list = [fake_host]
        mock_get_all.return_value = reserved_host_object_list
        mock_host_obj.return_value = fake_host

        reserved_host_list = [host.name for host in
                              reserved_host_object_list]

        notification = self._get_compute_host_type_notification()
        mock_notification_get.return_value = notification
        mock_host_failure.side_effect = self._fake_notification_workflow()

        self.engine.process_notification(self.context,
                                         notification=notification)

        update_data_by_host_failure = {
            'on_maintenance': True,
        }
        mock_host_update.assert_called_once_with(update_data_by_host_failure)
        mock_host_save.assert_called_once()
        self.assertEqual("finished", notification.status)
        mock_host_failure.assert_called_once_with(
            self.context,
            fake_host.name, fake_host.failover_segment.recovery_method,
            notification.notification_uuid,
            reserved_host_list=reserved_host_list,
            update_host_method=manager.update_host_method)
        mock_get_all.assert_called_once_with(self.context, filters={
            'failover_segment_id': fake_host.failover_segment.uuid,
            'reserved': True, 'on_maintenance': False})
        action = fields.EventNotificationAction.NOTIFICATION_PROCESS
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, notification, action=action,
                      phase=phase_start),
            mock.call(self.context, notification, action=action,
                      phase=phase_end)]
        mock_notify_about_notification_update.assert_has_calls(notify_calls)

    @mock.patch.object(host_obj.Host, "get_by_uuid")
    @mock.patch.object(host_obj.Host, "save")
    @mock.patch.object(host_obj.Host, "update")
    @mock.patch.object(host_obj.HostList, "get_all")
    @mock.patch("masakari.engine.drivers.taskflow."
                "TaskFlowDriver.execute_host_failure")
    @mock.patch.object(notification_obj.Notification, "save")
    @mock.patch.object(engine_utils, 'notify_about_notification_update')
    def test_process_notification_reserved_host_failure(
            self, mock_notify_about_notification_update,
            mock_notification_save, mock_host_failure, mock_get_all,
            mock_host_update, mock_host_save, mock_host_obj,
            mock_notification_get):
        fake_host = fakes.create_fake_host(reserved=True)
        fake_host.failover_segment = fakes.create_fake_failover_segment(
            recovery_method='reserved_host')
        reserved_host_object_list = [fake_host]
        mock_get_all.return_value = reserved_host_object_list
        mock_host_obj.return_value = fake_host

        notification = self._get_compute_host_type_notification()
        mock_notification_get.return_value = notification
        mock_host_failure.side_effect = self._fake_notification_workflow()

        reserved_host_list = [host.name for host in
                              reserved_host_object_list]

        self.engine.process_notification(self.context,
                                         notification=notification)

        update_data_by_host_failure = {
            'on_maintenance': True,
            'reserved': False,
        }
        mock_host_update.assert_called_once_with(update_data_by_host_failure)
        mock_host_save.assert_called_once()
        self.assertEqual("finished", notification.status)
        mock_host_failure.assert_called_once_with(
            self.context,
            fake_host.name, fake_host.failover_segment.recovery_method,
            notification.notification_uuid,
            reserved_host_list=reserved_host_list,
            update_host_method=manager.update_host_method)
        action = fields.EventNotificationAction.NOTIFICATION_PROCESS
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, notification, action=action,
                      phase=phase_start),
            mock.call(self.context, notification, action=action,
                      phase=phase_end)]
        mock_notify_about_notification_update.assert_has_calls(notify_calls)

    @mock.patch.object(host_obj.Host, "get_by_uuid")
    @mock.patch.object(host_obj.Host, "save")
    @mock.patch.object(host_obj.Host, "update")
    @mock.patch.object(host_obj.HostList, "get_all")
    @mock.patch("masakari.engine.drivers.taskflow."
                "TaskFlowDriver.execute_host_failure")
    @mock.patch.object(notification_obj.Notification, "save")
    @mock.patch.object(engine_utils, 'notify_about_notification_update')
    @mock.patch('traceback.format_exc')
    def test_process_notification_type_compute_host_recovery_exception(
            self, mock_format, mock_notify_about_notification_update,
            mock_notification_save, mock_host_failure, mock_get_all,
            mock_host_update, mock_host_save, mock_host_obj,
            mock_notification_get):
        mock_format.return_value = mock.ANY
        notification = self._get_compute_host_type_notification()
        mock_notification_get.return_value = notification
        fake_host = fakes.create_fake_host()
        mock_get_all.return_value = None
        fake_host.failover_segment = fakes.create_fake_failover_segment()
        mock_host_obj.return_value = fake_host
        mock_host_failure.side_effect = self._fake_notification_workflow(
            exc=exception.HostRecoveryFailureException)
        self.engine.process_notification(self.context,
                                         notification=notification)

        update_data_by_host_failure = {
            'on_maintenance': True,
        }
        mock_host_update.assert_called_once_with(update_data_by_host_failure)
        mock_host_save.assert_called_once()
        self.assertEqual("error", notification.status)
        e = exception.HostRecoveryFailureException('Failed to execute host '
                                                   'recovery.')
        action = fields.EventNotificationAction.NOTIFICATION_PROCESS
        phase_start = fields.EventNotificationPhase.START
        phase_error = fields.EventNotificationPhase.ERROR
        notify_calls = [
            mock.call(self.context, notification, action=action,
                      phase=phase_start),
            mock.call(self.context, notification, action=action,
                      phase=phase_error,
                      exception=str(e),
                      tb=mock.ANY)]
        mock_notify_about_notification_update.assert_has_calls(notify_calls)

    @mock.patch.object(host_obj.Host, "get_by_uuid")
    @mock.patch.object(host_obj.Host, "save")
    @mock.patch.object(host_obj.Host, "update")
    @mock.patch.object(host_obj.HostList, "get_all")
    @mock.patch("masakari.engine.drivers.taskflow."
                "TaskFlowDriver.execute_host_failure")
    @mock.patch.object(notification_obj.Notification, "save")
    @mock.patch.object(engine_utils, 'notify_about_notification_update')
    @mock.patch('traceback.format_exc')
    def test_process_notification_type_compute_host_skip_host_recovery(
            self, mock_format, mock_notify_about_notification_update,
            mock_notification_save, mock_host_failure, mock_get_all,
            mock_host_update, mock_host_save, mock_host_obj,
            mock_notification_get):
        mock_format.return_value = mock.ANY
        notification = self._get_compute_host_type_notification()
        mock_notification_get.return_value = notification
        fake_host = fakes.create_fake_host()
        mock_get_all.return_value = None
        fake_host.failover_segment = fakes.create_fake_failover_segment()
        mock_host_obj.return_value = fake_host
        # mock_host_failure.side_effect = str(e)
        mock_host_failure.side_effect = self._fake_notification_workflow(
            exc=exception.SkipHostRecoveryException)
        self.engine.process_notification(self.context,
                                         notification=notification)

        update_data_by_host_failure = {
            'on_maintenance': True,
        }
        mock_host_update.assert_called_once_with(update_data_by_host_failure)
        mock_host_save.assert_called_once()
        self.assertEqual("finished", notification.status)
        action = fields.EventNotificationAction.NOTIFICATION_PROCESS
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, notification, action=action,
                      phase=phase_start),
            mock.call(self.context, notification, action=action,
                      phase=phase_end)]
        mock_notify_about_notification_update.assert_has_calls(notify_calls)

    @mock.patch.object(notification_obj.Notification, "save")
    @mock.patch.object(engine_utils, 'notify_about_notification_update')
    @mock.patch("masakari.engine.drivers.taskflow."
                "TaskFlowDriver.execute_host_failure")
    def test_process_notification_type_compute_host_event_started(
            self, mock_host_failure, mock_notify_about_notification_update,
            mock_notification_save, mock_notification_get):
        notification = self._get_compute_host_type_notification()
        mock_notification_get.return_value = notification
        notification.payload['event'] = 'started'
        self.engine.process_notification(self.context,
                                         notification=notification)
        self.assertEqual("finished", notification.status)
        self.assertFalse(mock_host_failure.called)
        action = fields.EventNotificationAction.NOTIFICATION_PROCESS
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, notification, action=action,
                      phase=phase_start),
            mock.call(self.context, notification, action=action,
                      phase=phase_end)]
        mock_notify_about_notification_update.assert_has_calls(notify_calls)

    @mock.patch.object(notification_obj.Notification, "save")
    @mock.patch.object(engine_utils, 'notify_about_notification_update')
    @mock.patch("masakari.engine.drivers.taskflow."
                "TaskFlowDriver.execute_host_failure")
    def test_process_notification_type_compute_host_event_other(
            self, mock_host_failure, mock_notify_about_notification_update,
            mock_notification_save, mock_notification_get):
        notification = self._get_compute_host_type_notification()
        mock_notification_get.return_value = notification
        notification.payload['event'] = 'other'
        self.engine.process_notification(self.context,
                                         notification=notification)
        self.assertEqual("ignored", notification.status)
        self.assertFalse(mock_host_failure.called)
        action = fields.EventNotificationAction.NOTIFICATION_PROCESS
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, notification, action=action,
                      phase=phase_start),
            mock.call(self.context, notification, action=action,
                      phase=phase_end)]
        mock_notify_about_notification_update.assert_has_calls(notify_calls)

    @mock.patch("masakari.compute.nova.API.stop_server")
    @mock.patch.object(notification_obj.Notification, "save")
    @mock.patch.object(engine_utils, 'notify_about_notification_update')
    @mock.patch('traceback.format_exc')
    @mock.patch("masakari.compute.nova.API.get_server")
    def test_process_notification_type_vm_ignore_instance_in_paused(
            self, mock_get_server, mock_format,
            mock_notify_about_notification_update, mock_notification_save,
            mock_stop_server, mock_notification_get):
        mock_format.return_value = mock.ANY
        notification = _get_vm_type_notification()
        mock_notification_get.return_value = notification
        mock_get_server.return_value = fakes.FakeNovaClient.Server(
            id=1, uuid=uuidsentinel.fake_ins, host='fake_host',
            vm_state='paused', ha_enabled=True)

        self.engine.process_notification(self.context,
                                         notification=notification)
        self.assertEqual("ignored", notification.status)
        self.assertFalse(mock_stop_server.called)
        msg = ("Recovery of instance '%(instance_uuid)s' is ignored as it is "
               "in '%(vm_state)s' state.") % {'instance_uuid':
                uuidsentinel.fake_ins, 'vm_state': 'paused'}
        e = exception.IgnoreInstanceRecoveryException(msg)
        action = fields.EventNotificationAction.NOTIFICATION_PROCESS
        phase_start = fields.EventNotificationPhase.START
        phase_error = fields.EventNotificationPhase.ERROR
        notify_calls = [
            mock.call(self.context, notification, action=action,
                      phase=phase_start),
            mock.call(self.context, notification, action=action,
                      phase=phase_error,
                      exception=str(e),
                      tb=mock.ANY)]
        mock_notify_about_notification_update.assert_has_calls(notify_calls)

    @mock.patch("masakari.compute.nova.API.stop_server")
    @mock.patch.object(notification_obj.Notification, "save")
    @mock.patch.object(engine_utils, 'notify_about_notification_update')
    @mock.patch('traceback.format_exc')
    @mock.patch("masakari.compute.nova.API.get_server")
    def test_process_notification_type_vm_ignore_instance_in_rescued(
            self, mock_get_server, mock_format,
            mock_notify_about_notification_update, mock_notification_save,
            mock_stop_server, mock_notification_get):
        mock_format.return_value = mock.ANY
        notification = _get_vm_type_notification()
        mock_notification_get.return_value = notification
        mock_get_server.return_value = fakes.FakeNovaClient.Server(
            id=1, uuid=uuidsentinel.fake_ins, host='fake_host',
            vm_state='rescued', ha_enabled=True)

        self.engine.process_notification(self.context,
                                         notification=notification)
        self.assertEqual("ignored", notification.status)
        self.assertFalse(mock_stop_server.called)
        msg = ("Recovery of instance '%(instance_uuid)s' is ignored as it is "
               "in '%(vm_state)s' state.") % {'instance_uuid':
                uuidsentinel.fake_ins, 'vm_state': 'rescued'}
        e = exception.IgnoreInstanceRecoveryException(msg)
        action = fields.EventNotificationAction.NOTIFICATION_PROCESS
        phase_start = fields.EventNotificationPhase.START
        phase_error = fields.EventNotificationPhase.ERROR
        notify_calls = [
            mock.call(self.context, notification, action=action,
                      phase=phase_start),
            mock.call(self.context, notification, action=action,
                      phase=phase_error,
                      exception=str(e),
                      tb=mock.ANY)]
        mock_notify_about_notification_update.assert_has_calls(notify_calls)

    def test_process_notification_stop_from_recovery_failure(self,
                                                     mock_get_noti):
        noti_new = _get_vm_type_notification()
        mock_get_noti.return_value = _get_vm_type_notification(
            status="failed")

        with mock.patch("masakari.engine.manager.LOG.warning") as mock_log:
            self.engine.process_notification(self.context,
                                             notification=noti_new)
            mock_log.assert_called_once()
            args = mock_log.call_args[0]
            expected_log = ("Processing of notification is skipped to avoid "
                            "recovering from failure twice. "
                            "Notification received is '%(uuid)s' "
                            "and it's status is '%(new_status)s' and the "
                            "current status of same notification in db "
                            "is '%(old_status)s'")
            expected_log_args_1 = {'uuid': noti_new.notification_uuid,
                                   'new_status': noti_new.status,
                                   'old_status': "failed"}

            self.assertEqual(expected_log, args[0])
            self.assertEqual(expected_log_args_1, args[1])

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.host_failure.'
                'DisableComputeServiceTask.execute')
    @mock.patch('masakari.engine.drivers.taskflow.host_failure.'
                'PrepareHAEnabledInstancesTask.execute')
    @mock.patch('masakari.engine.drivers.taskflow.host_failure.'
                'EvacuateInstancesTask.execute')
    @mock.patch('masakari.engine.drivers.taskflow.no_op.LOG')
    def test_host_failure_custom_flow_for_auto_recovery(
            self, _mock_log, _mock_task1, _mock_task2, _mock_task3,
            _mock_novaclient, _mock_notification_get):
        # For testing purpose setting BACKEND as memory
        masakari.engine.drivers.taskflow.base.PERSISTENCE_BACKEND = 'memory://'
        self.override_config(
            "host_auto_failure_recovery_tasks",
            {'pre': ['disable_compute_service_task', 'no_op'],
             'main': ['prepare_HA_enabled_instances_task'],
             'post': ['evacuate_instances_task']},
            "taskflow_driver_recovery_flows")

        expected_msg_format = "Custom task executed successfully..!!"

        self.engine.driver.execute_host_failure(
            self.context, 'fake_host',
            fields.FailoverSegmentRecoveryMethod.AUTO,
            uuidsentinel.fake_notification)
        # Ensure custom_task added to the 'host_auto_failure_recovery_tasks'
        # is executed.
        _mock_log.info.assert_called_with(expected_msg_format)

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch.object(nova.API, "enable_disable_service")
    @mock.patch('masakari.engine.drivers.taskflow.host_failure.'
                'PrepareHAEnabledInstancesTask.execute')
    @mock.patch('masakari.engine.drivers.taskflow.host_failure.'
                'EvacuateInstancesTask.execute')
    @mock.patch('masakari.engine.drivers.taskflow.base.MasakariTask.'
                'update_details')
    @mock.patch('masakari.engine.drivers.taskflow.host_failure.LOG')
    def test_host_failure_flow_for_auto_recovery(self, _mock_log,
                                                 _mock_notify,
                                                 _mock_novaclient,
                                                 _mock_enable_disable,
                                                 _mock_task2, _mock_task3,
                                                 _mock_notification_get):
        self.novaclient = nova.API()
        self.fake_client = fakes.FakeNovaClient()
        self.override_config("wait_period_after_evacuation", 2)
        self.override_config("wait_period_after_service_update", 2)
        self.override_config("evacuate_all_instances",
                             True, "host_failure")

        _mock_novaclient.return_value = self.fake_client

        # create test data
        self.fake_client.servers.create(id="1", host="fake-host",
                                        ha_enabled=True)
        self.fake_client.servers.create(id="2", host="fake-host")

        instance_uuid_list = []
        for instance in self.fake_client.servers.list():
            instance_uuid_list.append(instance.id)

        instance_list = {
            "instance_list": ','.join(instance_uuid_list),
        }
        _mock_task2.return_value = instance_list

        # For testing purpose setting BACKEND as memory
        masakari.engine.drivers.taskflow.base.PERSISTENCE_BACKEND = 'memory://'

        self.engine.driver.execute_host_failure(
            self.context, "fake-host",
            fields.FailoverSegmentRecoveryMethod.AUTO,
            uuidsentinel.fake_notification)

        # make sure instance is active and has different host
        for server in instance_uuid_list:
            instance = self.novaclient.get_server(self.context, server)

            if CONF.host_failure.ignore_instances_in_error_state and getattr(
                    instance, 'OS-EXT-STS:vm_state') == 'error':
                self.assertEqual(
                    "fake-host", getattr(
                        instance, 'OS-EXT-SRV-ATTR:hypervisor_hostname'))
            else:
                self.assertNotEqual(
                    "fake-host", getattr(
                        instance, 'OS-EXT-SRV-ATTR:hypervisor_hostname'))

        # verify progress details
        _mock_notify.assert_has_calls([
            mock.call("Disabling compute service on host: 'fake-host'"),
            mock.call("Disabled compute service on host: 'fake-host'", 1.0)
        ])

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.host_failure.'
                'DisableComputeServiceTask.execute')
    @mock.patch('masakari.engine.drivers.taskflow.host_failure.'
                'PrepareHAEnabledInstancesTask.execute')
    @mock.patch('masakari.engine.drivers.taskflow.host_failure.'
                'EvacuateInstancesTask.execute')
    @mock.patch('masakari.engine.drivers.taskflow.no_op.LOG')
    def test_host_failure_custom_flow_for_rh_recovery(self, _mock_log,
                                                      _mock_task1,
                                                      _mock_task2,
                                                      _mock_task3,
            _mock_novaclient, _mock_notification_get):
        # For testing purpose setting BACKEND as memory
        masakari.engine.drivers.taskflow.base.PERSISTENCE_BACKEND = 'memory://'
        self.override_config(
            "host_rh_failure_recovery_tasks",
            {'pre': ['disable_compute_service_task'],
             'main': [],
             'post': ['no_op']},
            "taskflow_driver_recovery_flows")

        expected_msg_format = "Custom task executed successfully..!!"

        self.engine.driver.execute_host_failure(
            self.context, 'fake_host',
            fields.FailoverSegmentRecoveryMethod.RESERVED_HOST,
            uuidsentinel.fake_notification,
            update_host_method=manager.update_host_method,
            reserved_host_list=['host-1', 'host-2'])
        # Ensure custom_task added to the 'host_rh_failure_recovery_tasks'
        # is executed.
        _mock_log.info.assert_called_with(expected_msg_format)

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch.object(nova.API, "enable_disable_service")
    @mock.patch('masakari.engine.drivers.taskflow.host_failure.'
                'PrepareHAEnabledInstancesTask.execute')
    @mock.patch('masakari.engine.drivers.taskflow.host_failure.'
                'EvacuateInstancesTask.execute')
    @mock.patch('masakari.engine.drivers.taskflow.base.MasakariTask.'
                'update_details')
    @mock.patch('masakari.engine.drivers.taskflow.host_failure.LOG')
    def test_host_failure_flow_for_rh_recovery(self, _mock_log, _mock_notify,
                                               _mock_novaclient,
                                               _mock_enable_disable,
                                               _mock_task2, _mock_task3,
                                               _mock_notification_get):
        self.novaclient = nova.API()
        self.fake_client = fakes.FakeNovaClient()
        self.override_config("wait_period_after_evacuation", 2)
        self.override_config("wait_period_after_service_update", 2)
        self.override_config("evacuate_all_instances",
                             True, "host_failure")

        _mock_novaclient.return_value = self.fake_client

        # create test data
        self.fake_client.servers.create(id="1", host="fake-host",
                                        ha_enabled=True)
        self.fake_client.servers.create(id="2", host="fake-host")

        instance_uuid_list = []
        for instance in self.fake_client.servers.list():
            instance_uuid_list.append(instance.id)

        instance_list = {
            "instance_list": ','.join(instance_uuid_list),
        }
        _mock_task2.return_value = instance_list

        # For testing purpose setting BACKEND as memory
        masakari.engine.drivers.taskflow.base.PERSISTENCE_BACKEND = 'memory://'

        self.engine.driver.execute_host_failure(
            self.context, "fake-host",
            fields.FailoverSegmentRecoveryMethod.RESERVED_HOST,
            uuidsentinel.fake_notification,
            update_host_method=manager.update_host_method,
            reserved_host_list=['host-1', 'host-2'])

        # make sure instance is active and has different host
        for server in instance_uuid_list:
            instance = self.novaclient.get_server(self.context, server)
            self.assertNotEqual(
                "fake-host", getattr(
                    instance, 'OS-EXT-SRV-ATTR:hypervisor_hostname'))

        # verify progress details
        _mock_notify.assert_has_calls([
            mock.call("Disabling compute service on host: 'fake-host'"),
            mock.call("Disabled compute service on host: 'fake-host'", 1.0)
        ])

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.instance_failure.'
                'StopInstanceTask.execute')
    @mock.patch('masakari.engine.drivers.taskflow.instance_failure.'
                'StartInstanceTask.execute')
    @mock.patch('masakari.engine.drivers.taskflow.instance_failure.'
                'ConfirmInstanceActiveTask.execute')
    @mock.patch('masakari.engine.drivers.taskflow.no_op.LOG')
    def test_instance_failure_custom_flow_recovery(
            self, _mock_log, _mock_task1, _mock_task2, _mock_task3,
            _mock_novaclient, _mock_notification_get):
        # For testing purpose setting BACKEND as memory
        masakari.engine.drivers.taskflow.base.PERSISTENCE_BACKEND = 'memory://'
        self.override_config(
            "instance_failure_recovery_tasks",
            {'pre': ['stop_instance_task', 'no_op'],
             'main': ['start_instance_task'],
             'post': ['confirm_instance_active_task']},
            "taskflow_driver_recovery_flows")

        expected_msg_format = "Custom task executed successfully..!!"

        self.engine.driver.execute_instance_failure(
            self.context, uuidsentinel.fake_ins,
            uuidsentinel.fake_notification)
        # Ensure custom_task added to the 'instance_failure_recovery_tasks'
        # is executed.
        _mock_log.info.assert_called_with(expected_msg_format)

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.base.MasakariTask.'
                'update_details')
    @mock.patch('masakari.engine.drivers.taskflow.instance_failure.LOG')
    def test_instance_failure_flow_recovery(self, _mock_log, _mock_notify,
                                            _mock_novaclient,
                                            _mock_notification_get):
        self.novaclient = nova.API()
        self.fake_client = fakes.FakeNovaClient()
        self.override_config('wait_period_after_power_off', 2)
        self.override_config('wait_period_after_power_on', 2)
        instance_id = uuidsentinel.fake_ins

        _mock_novaclient.return_value = self.fake_client

        # create test data
        self.fake_client.servers.create(instance_id,
                                        host="fake-host",
                                        ha_enabled=True)

        # For testing purpose setting BACKEND as memory
        masakari.engine.drivers.taskflow.base.PERSISTENCE_BACKEND = 'memory://'

        self.engine.driver.execute_instance_failure(
            self.context, instance_id,
            uuidsentinel.fake_notification)

        # verify instance is in active state
        instance = self.novaclient.get_server(self.context, instance_id)
        self.assertEqual('active',
                         getattr(instance, 'OS-EXT-STS:vm_state'))

        _mock_notify.assert_has_calls([
            mock.call('Stopping instance: ' + instance_id),
            mock.call("Stopped instance: '" + instance_id + "'", 1.0),
            mock.call("Starting instance: '" + instance_id + "'"),
            mock.call("Instance started: '" + instance_id + "'", 1.0),
            mock.call("Confirming instance '" + instance_id +
                      "' vm_state is ACTIVE"),
            mock.call("Confirmed instance '" + instance_id +
                      "' vm_state is ACTIVE", 1.0)
        ])

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.process_failure.'
                'DisableComputeNodeTask.execute')
    @mock.patch('masakari.engine.drivers.taskflow.process_failure.'
                'ConfirmComputeNodeDisabledTask.execute')
    @mock.patch('masakari.engine.drivers.taskflow.no_op.LOG')
    def test_process_failure_custom_flow_recovery(
            self, _mock_log, _mock_task1, _mock_task2, _mock_novaclient,
            _mock_notification_get):
        # For testing purpose setting BACKEND as memory
        masakari.engine.drivers.taskflow.base.PERSISTENCE_BACKEND = 'memory://'
        self.override_config(
            "process_failure_recovery_tasks",
            {'pre': ['disable_compute_node_task', 'no_op'],
             'main': ['confirm_compute_node_disabled_task'],
             'post': []},
            "taskflow_driver_recovery_flows")

        expected_msg_format = "Custom task executed successfully..!!"

        self.engine.driver.execute_process_failure(
            self.context, 'nova-compute', 'fake_host',
            uuidsentinel.fake_notification, )

        _mock_log.info.assert_any_call(expected_msg_format)
        # Ensure custom_task added to the 'process_failure_recovery_tasks'
        # is executed.
        _mock_log.info.assert_called_with(expected_msg_format)

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.base.MasakariTask.'
                'update_details')
    @mock.patch('masakari.engine.drivers.taskflow.process_failure.LOG')
    def test_process_failure_flow_recovery(self, _mock_log, _mock_notify,
                                           _mock_novaclient,
                                           _mock_notification_get):
        self.novaclient = nova.API()
        self.fake_client = fakes.FakeNovaClient()
        _mock_novaclient.return_value = self.fake_client

        # create test data
        self.fake_client.services.create("1", host="fake-host",
                                         binary="nova-compute",
                                         status="enabled")

        # For testing purpose setting BACKEND as memory
        masakari.engine.drivers.taskflow.base.PERSISTENCE_BACKEND = 'memory://'

        self.engine.driver.execute_process_failure(
            self.context, "nova-compute", "fake-host",
            uuidsentinel.fake_notification)

        # verify service is disabled
        self.assertTrue(self.novaclient.is_service_down(self.context,
                                                        "fake-host",
                                                        "nova-compute"))
        # verify progress details
        _mock_notify.assert_has_calls([
            mock.call("Disabling compute service on host: 'fake-host'"),
            mock.call("Disabled compute service on host: 'fake-host'", 1.0),
            mock.call("Confirming compute service is disabled on host: "
                      "'fake-host'"),
            mock.call("Confirmed compute service is disabled on host: "
                      "'fake-host'", 1.0)
        ])

    @mock.patch.object(notification_obj.Notification, "save")
    @mock.patch('masakari.engine.drivers.taskflow.driver.TaskFlowDriver.'
                'get_notification_recovery_workflow_details')
    def test_get_notification_recovery_workflow_details(self,
                                                        mock_progress_details,
                                                        mock_save,
                                                        mock_notification_get):
        notification = fakes.create_fake_notification(
            type="VM", id=1, payload={
                'event': 'fake_event', 'instance_uuid': uuidsentinel.fake_ins,
                'vir_domain_event': 'fake_vir_domain_event'
            },
            source_host_uuid=uuidsentinel.fake_host,
            generated_time=NOW, status="new",
            notification_uuid=uuidsentinel.fake_notification,)

        mock_notification_get.return_value = notification
        self.engine.driver.get_notification_recovery_workflow_details(
            self.context, notification)

        mock_progress_details.assert_called_once_with(
            self.context, notification)



# Copyright 2016 NTT DATA.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from masakari.api.validation import parameter_types


_base = {
    'type': 'object',
    'properties': {
        'host': {
            'type': 'object',
            'properties': {
                'name': parameter_types.name,
                'type': parameter_types.type,
                'control_attributes': parameter_types.description,
                'reserved': parameter_types.boolean,
                'on_maintenance': parameter_types.boolean
            },
            'additionalProperties': False
        }
    },
    'required': ['host'],
    'additionalProperties': False
}


create = copy.deepcopy(_base)
create['properties']['host']['required'] = ['name', 'type',
                                            'control_attributes']


update = copy.deepcopy(_base)
update['properties']['host']['anyOf'] = [{'required': ['name']},
                                         {'required': ['type']},
                                         {'required': ['control_attributes']},
                                         {'required': ['reserved']},
                                         {'required': ['on_maintenance']},
                                         ]
